﻿#region using
using DataDynamics.ActiveReports;
using DataDynamics.ActiveReports.Export.Pdf;
using Kercia.AlphaVote.Utils;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
#endregion

namespace UtilsActiveReport
{
    /// <summary>
    /// Classe de base pour tous les documents ActiveReport de l'application.
    /// </summary>
    public class Report:ActiveReport
    {
        #region Membres Privés

        private static PdfExport _PdfExport;
        private static PdfExport _PdfExportHD;

        //Index de la ligne en cours de processing
        private int _IndexLigneCourante = 0;

        //Collection des subreports de ce report
        private IList<SubReport> _SubReports;

        //Gestion progression de la génération
        private decimal _ProgressionGeneration;
        public event ProgressChangedEventHandler OnProgressChanged;
        public delegate void ProgressChangedEventHandler(object sender, EventArgs e);

        #endregion

        #region Constructeurs

        public Report():base()
        {
            _ProgressionGeneration = 0;
            this.Document.CacheToDisk = false;
        }

        #endregion

        #region Membres Publics

        /// <summary>
        /// Verrou à utiliser IMPERATIVEMENT lors de l'appel d'une méthode d'export PDF depuis un document Active Report, afin d'éviter tout problème de génération parallèle.
        /// Ce fonctionnement est nécessaire car l'objet "PdfExport" est instancié une seule fois, car son instanciation peut prendre du temps (quelques secondes selon la machine), et il n'est donc pas raisonnable de l'instancier pour chaque export de document.
        /// </summary>
        public static object VerrouExportPdf = new object();

        /// <summary>
        /// Export pdf standard à utiliser pour la génération des documents.
        /// </summary>
        public static PdfExport PdfExport 
        {
            get 
            {
                if (_PdfExport == null) 
                {
                    _PdfExport = new PdfExport();
                    _PdfExport.NeverEmbedFonts = string.Empty;
                }

                return _PdfExport;
            }
        }

        /// <summary>
        /// Export pdf à utiliser pour la génération de documents avec des images d'un haut niveau de détails.
        /// A éviter pour des documents utilisant un grand nombre d'images ou des images de grande taille car le fichier généré pèsera significativement plus lourd.
        /// </summary>
        public static PdfExport PdfExportHD 
        {
            get 
            {
                if (_PdfExportHD == null) 
                {
                    _PdfExportHD = new PdfExport();
                    _PdfExportHD.NeverEmbedFonts = string.Empty;
                    _PdfExportHD.ImageQuality = ImageQuality.Highest;
                }

                return _PdfExportHD;
            }
        }

        #endregion

        /// <summary>
        /// Index de la ligne en cours de processing
        /// </summary>
        protected int IndexLigneCourante 
        {
            get { return _IndexLigneCourante; }
            set { _IndexLigneCourante = value; }
        }

        /// <summary>
        /// Progression de la génération du report.
        /// </summary>
        public decimal ProgressionGeneration 
        {
            get { return _ProgressionGeneration; }
            protected set 
            { 
                _ProgressionGeneration = value;
                if(OnProgressChanged != null)
                    OnProgressChanged(this, new EventArgs());
            }
        }

        #region Gestion des Subreports

        /// <summary>
        /// Enregistre un subreport dans ce report.
        /// Nécessaire afin d'avoir une bonne gestion des ressources.
        /// </summary>
        /// <param name="subReport">Subreport à enregistrer</param>
        public void EnregistrerSubReport(SubReport subReport) 
        {
            try
            {
                if (subReport == null)
                    return;

                if (subReport.Report == null)
                    throw new Exception(Name + " - Le subreport " + subReport.Name + " n'a aucun report défini");

                if (_SubReports == null)
                    _SubReports = new List<SubReport>();

                LoggerUniversel.Debug(Name + " - Enregistrement du subreport " + subReport.Name);

                _SubReports.Add(subReport);
            }
            catch (Exception e) 
            {
                LoggerUniversel.Error("Erreur lors de l'enregistrement du subreport",e);
                throw;
            }
        }

        /// <summary>
        /// Indique si un subreport est déjà défini.
        /// Il est important d'effectuer ce test afin d'éviter d'affecter inutilement des subreports.
        /// </summary>
        /// <param name="subReport">Subreport concerné</param>
        /// <returns></returns>
        public bool SubReportDefini(SubReport subReport)
        {
            if (subReport==null || _SubReports==null || _SubReports.Count==0)
                return false;

            //Recherche du composant de nom donné
            foreach (SubReport sr in _SubReports)
                if (sr == subReport)
                    return true;
            
            return false;
        }

        /// <summary>
        /// Libère les ressources occupées par les subreports.
        /// </summary>
        public void DisposeSubReports() 
        {
            LoggerUniversel.Debug(Name + " - Libération des ressources occupées par les subreports");

            if (_SubReports == null||_SubReports.Count==0)
                return;

            while(_SubReports.Count>0)
                DisposeSubReport(_SubReports[0]);

            //Réinitialisation des listes
            _SubReports = null;
        }

        /// <summary>
        /// Libère les ressources occupées par un subreport.
        /// </summary>
        /// <param name="subreport">Subreport concerné</param>
        private void DisposeSubReport(SubReport subReport) 
        {
            if (subReport == null)
                return;

            LoggerUniversel.Debug(Name + " - Libération des ressources occupées par le subreport " + subReport.Name);

            //Retrait de la liste des subreports
            if (_SubReports != null)
                _SubReports.Remove(subReport);

            subReport.Report.Document.Dispose();
            subReport.Report.Dispose();
            //subReport.Report = null;
        }

        #endregion

        #region Méthodes

        /// <summary>
        /// Eclate une chaine de façon à ce qu'elle puisse etre imprimée sur le cerfa de façon à ce que chaque caractère entre dans une case préimprimée.
        /// </summary>
        /// <param name="chaine">Chaine à éclater</param>
        /// <returns></returns>
        protected string EclaterPourImpressionCases(string chaine, int nbEspaces)
        {
            if (string.IsNullOrEmpty(chaine))
                return chaine;

            char[] caraChaine = chaine.ToCharArray();
            StringBuilder strBldr = new StringBuilder();
            for (int i = 0; i < caraChaine.Length - 1; i++)
            {
                strBldr = strBldr.Append(caraChaine[i]);
                for (int j = 0; j < nbEspaces; j++)
                    strBldr = strBldr.Append(' ');
            }

            strBldr = strBldr.Append(caraChaine[caraChaine.Length - 1]);

            return strBldr.ToString();
        }

        protected SizeF GetSizeTextBox(TextBox txtBox) 
        {
            SizeF sizeText = new SizeF();
            Font old_font_page = this.CurrentPage.Font;
            this.CurrentPage.Font = txtBox.Font;
            sizeText = this.CurrentPage.MeasureText(txtBox.Text);

            this.CurrentPage.Font = old_font_page;

            return sizeText;
        }

        /// <summary>
        /// Centre en largeur les contrôles fournis dans le composant conteneur indiqué, selon que les contrôles sont visibles ou non.
        /// </summary>
        /// <param name="document"></param>
        /// <param name="controlesACentrer">Les contrôles à centrer, dans l'ordre séquentiel</param>
        protected void CentrerTableau(bool redimensionnerControles, List<ARControl> controlesACentrer) 
        {
            if (controlesACentrer == null || controlesACentrer.Count == 0)
                return;

            float largeurConteneur = this.PrintWidth;
            float totLargeurCourante = 0;
            int nbControlesVisibles = 0;
            foreach (ARControl controle in controlesACentrer)
                totLargeurCourante += controle.Width;

            float totLargeurControlesVisibles = 0;
            foreach (ARControl controle in controlesACentrer)
                if (controle.Visible)
                {
                    totLargeurControlesVisibles += controle.Width;
                    nbControlesVisibles++;
                }
                    
            float largeurARepartir = totLargeurCourante - totLargeurControlesVisibles;
            if (largeurARepartir == 0)
                return;

            float margeGauche = controlesACentrer[0].Location.X;
            float margeDroite = this.PrintWidth - (controlesACentrer[controlesACentrer.Count - 1].Location.X + controlesACentrer[controlesACentrer.Count - 1].Width);

            if (redimensionnerControles)
            {
                //Chaque contrôle visible est redimensionné afin d'exploiter au maximum la place

                //Largeur à ajouter pour chaque contrôle
                float largeurAjoutControle = (totLargeurCourante - (margeGauche + margeDroite) - totLargeurControlesVisibles) / nbControlesVisibles;
                totLargeurControlesVisibles += (nbControlesVisibles * largeurAjoutControle);

                //Positionnement du 1er contrôle
                int indexDernierControleVisible = -1;
                for (int i = 0; i < controlesACentrer.Count; i++)
                {
                    if(controlesACentrer[i].Visible)
                    {
                        if (indexDernierControleVisible == -1)
                        {
                            controlesACentrer[i].Location = new PointF((largeurConteneur - totLargeurControlesVisibles) / 2, controlesACentrer[i].Location.Y);
                            controlesACentrer[i].Width = controlesACentrer[i].Width + largeurAjoutControle;
                        }
                        else 
                        {
                            controlesACentrer[i].Location = new PointF(controlesACentrer[indexDernierControleVisible].Location.X + controlesACentrer[indexDernierControleVisible].Width, controlesACentrer[i].Location.Y);
                            controlesACentrer[i].Width = controlesACentrer[i].Width + largeurAjoutControle;
                        }
                        indexDernierControleVisible = i;
                    }
                }
            }
            else 
            {
                //La taille des contrôles visibles est conservée, mais leur position est réajustée afin qu'ils soient centrés sur la page

                //Positionnement du 1er contrôle
                int indexDernierControleVisible = -1;
                for (int i = 0; i < controlesACentrer.Count; i++) 
                {
                    if (controlesACentrer[i].Visible) 
                    {
                        if (indexDernierControleVisible == -1)
                            controlesACentrer[i].Location = new PointF((largeurConteneur - totLargeurControlesVisibles) / 2, controlesACentrer[i].Location.Y);
                        else 
                            controlesACentrer[i].Location = new PointF(controlesACentrer[indexDernierControleVisible].Location.X + controlesACentrer[indexDernierControleVisible].Width, controlesACentrer[i].Location.Y);
                        
                        indexDernierControleVisible = i;
                    }   
                }
            }
        }

        /// <summary>
        /// Ajuste la hauteur des cellules d'une ligne d'un tableau selon les textes de contenu.
        /// A utiliser dans l'événement BeforePrint.
        /// </summary>
        /// <param name="cellules"></param>
        protected void AjusterHauteurLigneTableau(IList<ARControl> cellules) 
        {
            if (cellules == null || cellules.Count == 0)
                return;

            float hauteurMax = 0;
            foreach (ARControl cellule in cellules)
                if (cellule.Visible && cellule.Height > hauteurMax)
                    hauteurMax = cellule.Height;

            foreach (ARControl cellule in cellules)
                cellule.Height = hauteurMax;
        }

        #endregion
    }
}
